
##
# This module providing methods to extend your ActiveRecord model to simplify use Audiomatic.
# If you want to use them, you'll need to include it in your model:
#
#    class MyModel < ApplicationRecord
#      include Neutrino::AudiomaticRails::Model
#    end
#
#And now you can customize your model.
#You have to show your model file to anlyze and set up what you want to analyze
#
#
# === Example:
#
#    class MyModel < ApplicationRecord
#      include Neutrino::AudiomaticRails::Model
#       audiomatic_file_url :my_url=
#       analyze :duraion, field: :duration
#    end
#
module Neutrino::AudiomaticRails
  module Model

    ProcessorException = Class.new(StandardError)

    def self.included(receiver)
      receiver.extend ClassMethods
      receiver.send :include, InstanceMethods 
      receiver.send :attr_accessor, :block_analyze
      receiver.send :validate, :valid_audiomatic
      receiver.after_save :mount_file
      receiver.after_save :send_to_audiomatic
    end

    module InstanceMethods
      To_file = [:waveform, :webbrowser]

      def store file 
        @cache_file = file      
      end

      def mount_file
        unless block_analyze
          unless @cache_file.nil?
            path = FileUtils.mkdir_p "#{Rails.root}/tmp/audio/#{self.class.to_s}/"
            file = File.new(path[0] + "#{self.id}", "w") 
            FileUtils.cp @cache_file.path, file.path
          end
        end
      end

      def default_file_url
        url = Rails.application.routes.url_helpers.send( "neutrino_audiomatic_file_#{ self.class.to_s.underscore }_url")
        url + "?id=#{self.id}"
      end  


      ##
      #Prepares and sands request to Audiomatic
      #
      def single_processor_request(processor)
        options = _parse_options(self.class.options)
        ::Neutrino::Audiomatic::Processor.send_request(
          ::Neutrino::Audiomatic::Processor.processor_url(processor), 
          options.merge(url: self.send(self.class.file_url)))
      end 

      ##
      #Sends all processors to Auidomatic. 
      #Works after save model.
      #
      def send_to_audiomatic
        processors = self.class.processors
        unless  block_analyze
          self.block_analyze = true
          answers = processors.keys.map { | processor | single_processor_request(processor) } 
          error_interpreter answers
        end
      end

      ##
      #Reacts on error during sending request to Audiomatic.
      #Two strategy:
      #  -Sets ActiveRecor errors and model is removed from database.
      #  -Replays sending request until successfull answer, and allows customize rection on error
      # === Example (seccond strategy, using Neutrino::AudiomaticRails.audiomatic_error)
      #
      #   audiomatic_error :audiomatic_error=
      #
      #     def audiomatic_error= errors
      #       errors.each do |error|
      #          Rails.logger.errors(error)
      #       end 
      #       true   
      #     end
      #
      ##
      def error_interpreter answers
        if self.class.user_error_solution.nil?
          answers.each { |ans| audiomatic_error ans[:message] if ans.include?(:error) }
          raise ActiveRecord::RecordInvalid.new(self) unless save
        else
          unless  answers.all? { |ans| ans.include?(:success)} 
            Neutrino::AudiomaticRails::SendToAudiomaticJob.set(wait: 5.second).perform_later ({id: self.id, klass: self.class.to_s})
            self.send((self.class.user_error_solution), answers) 
          end
        end
      end    
      
      def audiomatic_error message
        @audiomatic_error ||= []
        @audiomatic_error = message
      end 
      
      def valid_audiomatic
        Array.wrap(@audiomatic_error).each { |error| errors.add(:audiomatic_error, error) }
      end    

      def _parse_options(options)
        if self.class.callback_url.nil?
          {callback_url: Neutrino::AudiomaticRails.gen_callback_path(self)}
        else
          {callback_url: self.send(self.class.callback_url)}
        end
      end

      def _save_to answer
        ans_processor, value = convert_answer(answer) 
        to_field, processor, field, transformation = find_processor answer
        if to_field.nil? && !processor.nil?
          if To_file.include?(processor)
            save_file processor, value
          else
            self.send(self.class.save_to(processor), answer)
          end        
        else 
          if to_field
            transformation.each { |fun| value = self.send(fun, value) } unless transformation.nil?
            self.write_attribute(field.to_s, value)   
            true
          end 
        end   
      rescue Exception => e
        Rails.logger.error(e) and false
      end 

      def save_file processor, value
        require 'open-uri' 
        path = Rails.root.join('tmp','audiomatic')
        Dir.mkdir(path) unless path.exist?
        path = path.join("#{processor.to_s}#{Time.now.to_i}")
        _file = open(path, "wb") do |file|
          file <<  open(value).read
        end
        if self.class.mount_on(processor).nil?
          self.send(self.class.save_to(processor), _file)
        else
          self.send(self.class.mount_on(processor)).store! open(_file)
        end
      end     

      def convert_answer answer
        [answer['key'].intern, answer['value']]
      end

      def gen_callback_url(file)
        Neutrino::AudiomaticRails.route_for(file.class, file.id)
      end 

      def find_processor answer
        keys = self.class.processors.keys
        user_processor = self.class.processors  
        keys.each do |key|  
          if user_processor[key].include?(answer['key'].intern)
            case user_processor[key]
            when Array then
              field = answer['key'].intern
              transformation = nil
            when Hash then  
              field = user_processor[key][answer['key'].intern][:field] || key
              transformation = user_processor[key][answer['key'].intern][:transformation]
            end 
            return [true, key, field, transformation] 
          else
            if !user_processor[answer['key'].intern].nil? && [:save_to, :mount_on].include?( user_processor[ processor = answer['key'].intern].keys[0])
              return [nil, processor, nil, nil]
            end 
          end
        end
        if answer[/^replaygain.*/]
          [nil, :replaygain, nil, nil]
        else
          [nil, :tags , nil, nil]
        end
      end 
      ##
      #Allows update_attributes without sending request to Auiomatic. 
      #
      def update_attributes_manualy(attributes, option = {})
        self.block_analyze = true
        update_attributes(attributes, option)
      end 
      def update_manualy(attribute, value)
        self.block_analyze = true
        self[attribute] = value
        save
      end   
    end # IstanceMethod

    module ClassMethods

      Processors = [:duration, :replaygain, :tags, :webbrowser, :waveform]
      FieldOptionParams = [:field, :transformation]

      #Sets url of file which will be analyze
      def audiomatic_file_url method=nil
        if method.nil?
          @_file_url = default_file_url
        else
          @_file_url = method 
        end
      end

      def file_url
        if @_file_url.nil?
          :default_file_url
        else
          @_file_url
        end
      end

      def callback_url method=nil
        unless method.nil?
          @_callback_url = method
        else
          @_callback_url
        end
      end 

      def processors 
        @_processors
      end
      ##
      #Allows set customize rection on saving audiomatic response
      #
      # ===Example
      #
      #  analyze :duration, save_to: :save_to=
      #
      #     def save_to= result
      #       if result > 0     
      #         self.duration = result
      #         true
      #       else
      #         {error: "duration < 0"}
      #       end
      #     end
      #
      def save_to processor 
        @_processors[processor.intern][:save_to]
      end

      def mount_on processor
        @_processors[processor.intern][:mount_on]
      end
      ##
      #Saves result of analyze in ActiveRecord field 
      # 
      # ===Example
      #
      #   analyze :duration, field: :duration
      #
      def field processor
        @_processors[processor.intern][:field]
      end

      def user_error_solution
        @_user_error_solution
      end
      ##
      #Allows set customize rection on error duting sending request to Audiomatic, such as:
      #   => "Invalid credentials"
      #   => "Rate limit reached"
      #
      # ===Example
      #
      #  audiomatic_error :audiomatic_error
      #
      #     def audiomatic_error= result
      #       raise ActiveRecord::RecordInvalid.new(self)
      #     end
      #
      def audiomatic_error method
        @_user_error_solution =  method
      end  
      ##
      #Modifies response data from Audiomatic before saving. Works only if you use method field.
      #
      # ===Example - convert response duration from miliseconds to seconds
      #
      #  analyze :duration, {duration: {field: :duration, transformation: :to_seconds=} 
      #
      #     def to_seconds= result
      #       result.to_i / 1000
      #     end
      #
      def transformation processor
        @_processors[processor.intern][:transformation]
      end

      ##
      #Sets processor which will be done on the file. 
      #Gives posibillites to sets method which convert result.
      #
      # ===Example
      #
      #   analyze :duration, {duration:  {transformation: :to_ms=}}
      #   analyze :replay_gain, save_to: :save_replay_gain=
      #
      def analyze(processor, options = {})
        errors = validate_analyze(processor, options)
        raise ProcessorException.new(errors) unless errors.empty?

        @_processors ||= {}
        @_processors[processor] = options
      end

      def validate_analyze(processor, options = {})
        return ["Processor #{processor} is not valid. It should be one of #{Processors}"] unless Processors.include?(processor)
        return [] if options.include?(:save_to) || options.include?(:mount_on)
        case options
        when Array
          return ["Model #{self.class} hasn't got requiring field "] unless options.all? { |elem| field_exists?(elem) } # Change it to check fields or remove entirely
        when Hash 
          return ["Options of processor #{processor} are not valid"] unless options.values.all? { |suboptions| validate_analyze_options(suboptions) } # {field: asdfsdf, transforamtion: asdfdsf},  :transfr 
        else
          raise ProcessorException.new(["Options passed to #{self.class}#analyze - #{processor} invalid.  Should be hash or array, is #{options.class}"])
        end
        []
      end  

      def validate_analyze_options(suboptions)
        case suboptions
        when Symbol then  
          field_exists?(suboptions)
        when Hash 
          (suboptions[:field].nil? || field_exists?(suboptions[:field].to_s)) &&
          (suboptions[:transformation].nil? || suboptions[:transformation].kind_of?(Array)) &&
          suboptions.keys.all? { |elem| FieldOptionParams.include?(elem) }
        else
          raise ProcessorException.new(["Options passed to #{self.class} analyze - #{processor} invalid.  Should be hash or symbol, is #{suboptions.class}"])
        end
      end

      def field_exists?(field)
        self.column_names.include?(field.to_s)
      rescue
       false
     end

     def clean_processors
      @_processors = {}
    end  

    def options
      @_options
    end
    end # ClassMethod
  end
end # Neutrino::AudiomaticRails::Model
